package ch.zhaw.pm2.multichat.server;

import ch.zhaw.pm2.multichat.protocol.ChatProtocolException;
import ch.zhaw.pm2.multichat.protocol.ConnectionHandler;
import static ch.zhaw.pm2.multichat.protocol.ConnectionHandler.State.*;
import ch.zhaw.pm2.multichat.protocol.NetworkHandler;

import java.io.EOFException;
import java.io.IOException;
import java.net.SocketException;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Scanner;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;



public class ServerConnectionHandler extends ConnectionHandler implements Runnable{
    private static final AtomicInteger connectionCounter = new AtomicInteger(0);
    private final int connectionId = connectionCounter.incrementAndGet();
    private final Map<String,ServerConnectionHandler> connectionRegistry;

    private ReentrantLock mutex;

    private Condition nameComplete;

    private String userName = "Anonymous-"+connectionId;
    private State state = NEW;

    @Override
    public void run() {
        startReceiving();
    }

    public ServerConnectionHandler(NetworkHandler.NetworkConnection<String> connection,
                                   Map<String,ServerConnectionHandler> registry, ReentrantLock mutex, Condition nameComplete) {
        super();
        setConnection(connection);
        Objects.requireNonNull(connection, "Connection must not be null");
        Objects.requireNonNull(registry, "Registry must not be null");
        this.connectionRegistry = registry;
        this.mutex = mutex;
        this.nameComplete = nameComplete;
    }

    public String getUserName() {
        return this.userName;
    }

    private void startReceiving() {
            System.out.println("Starting Connection Handler for new User");
            try {
                System.out.println("Start receiving data...");
                while (getConnection().isAvailable()) {
                    String data = getConnection().receive();
                    processData(data);
                }
                System.out.println("Stopped recieving data");
            } catch (SocketException e) {
                System.out.println("Connection terminated locally");
                connectionRegistry.remove(userName);
                System.out.println("Unregistered because client connection terminated: " + userName + " " + e.getMessage());
            } catch (EOFException e) {
                System.out.println("Connection terminated by remote");
                connectionRegistry.remove(userName);
                System.out.println("Unregistered because client connection terminated: " + userName + " " + e.getMessage());
            } catch (IOException e) {
                System.err.println("Communication error: " + e);
            } catch (ClassNotFoundException e) {
                System.err.println("Received object of unknown type: " + e.getMessage());
            }
            System.out.println("Stopping Connection Handler for " + userName);

    }

    private void stopReceiving() {
        System.out.println("Closing Connection Handler for " + userName);
        try {
            System.out.println("Stop receiving data...");
            getConnection().close();
            System.out.println("Stopped receiving data.");
        } catch (IOException e) {
            System.err.println("Failed to close connection." + e);
        }
        System.out.println("Closed Connection Handler for " + userName);
    }

    private void processData(String data)  {
        try {
            // parse data content
            Scanner scanner = new Scanner(data);
            String sender = null;
            String reciever = null;
            String type = null;
            String payload = null;
            if (scanner.hasNextLine()) {
                sender = scanner.nextLine();
            } else {
                throw new ChatProtocolException("No Sender found");
            }
            if (scanner.hasNextLine()) {
                reciever = scanner.nextLine();
            } else {
                throw new ChatProtocolException("No Reciever found");
            }
            if (scanner.hasNextLine()) {
                type = scanner.nextLine();
            } else {
                throw new ChatProtocolException("No Type found");
            }
            if (scanner.hasNextLine()) {
                payload = scanner.nextLine();
            }

            // dispatch operation based on type parameter
            if (type.equals(getDataTypeConnect())) {
                if (this.state != NEW) throw new ChatProtocolException("Illegal state for connect request: " + state);
                if (sender == null || sender.isBlank()) sender = this.userName;
                if (connectionRegistry.containsKey(sender))
                    throw new ChatProtocolException("User name already taken: " + sender);
                mutex.lock();
                try {
                    this.userName = sender;
                    nameComplete.signal();
                }
                finally {
                    mutex.unlock();
                }
                connectionRegistry.put(userName, this);
                sendData(USER_NONE, userName, getDataTypeConfirm(), "Registration successfull for " + userName);
                this.state = CONNECTED;
            } else if (type.equals(getDataTypeConfirm())) {
                System.out.println("Not expecting to receive a CONFIRM request from client");
            } else if (type.equals(getDataTypeDisconnect())) {
                if (state == DISCONNECTED)
                    throw new ChatProtocolException("Illegal state for disconnect request: " + state);
                if (state == CONNECTED) {
                    connectionRegistry.remove(this.userName);
                }
                sendData(USER_NONE, userName, getDataTypeConfirm(), "Confirm disconnect of " + userName);
                this.state = DISCONNECTED;
                this.stopReceiving();
            } else if (type.equals(getDataTypeMessage())) {
                if (state != CONNECTED) throw new ChatProtocolException("Illegal state for message request: " + state);
                if (USER_ALL.equals(reciever)) {
                    for (ServerConnectionHandler handler : connectionRegistry.values()) {
                        handler.sendData(sender, reciever, type, payload);
                    }
                } else {
                    ServerConnectionHandler handler = connectionRegistry.get(reciever);
                    if (handler != null) {
                        handler.sendData(sender, reciever, type, payload);
                        if(!reciever.equals(sender)){
                            sendData(sender, reciever, type, payload); //send message to sender if it's a direct message and sender is not receiver.
                        }
                    } else {
                        this.sendData(USER_NONE, userName, getDataTypeError(), "Unknown User: " + reciever);
                    }
                }
            } else if (type.equals(getDataTypeError())) {
                System.err.println("Received error from client (" + sender + "): " + payload);
            } else {
                System.err.println("Unknown data type received: " + type);

            }
        } catch(ChatProtocolException e) {
            System.out.println("Error while processing data" + e.getMessage());
            sendData(USER_NONE, userName, getDataTypeError(), e.getMessage());
        }
    }

    private void sendData(String sender, String receiver, String type, String payload) {
        if (getConnection().isAvailable()) {
            new StringBuilder();
            String data = new StringBuilder()
                .append(sender+"\n")
                .append(receiver+"\n")
                .append(type+"\n")
                .append(payload+"\n")
                .toString();
            try {
                getConnection().send(data);
            } catch (SocketException e) {
                System.out.println("Connection closed: " + e.getMessage());
            } catch (EOFException e) {
                System.out.println("Connection terminated by remote");
            } catch(IOException e) {
                System.out.println("Communication error: " + e.getMessage());
            }
        }
    }
}
